/*
+-------------------------------- DISCLAIMER ---------------------------------+
|                                                                             |
| This application program is provided to you free of charge as an example.   |
| Despite the considerable efforts of Euresys personnel to create a usable    |
| example, you should not assume that this program is error-free or suitable  |
| for any purpose whatsoever.                                                 |
|                                                                             |
| EURESYS does not give any representation, warranty or undertaking that this |
| program is free of any defect or error or suitable for any purpose. EURESYS |
| shall not be liable, in contract, in torts or otherwise, for any damages,   |
| loss, costs, expenses or other claims for compensation, including those     |
| asserted by third parties, arising out of or in connection with the use of  |
| this program.                                                               |
|                                                                             |
+-----------------------------------------------------------------------------+
*/

// Helper for Virtual File System
// 2.0.0.15

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdexcept>
#include <memory>
#include <poll.h>
#include <sys/types.h>
#include <sys/mman.h>


#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/xmlwriter.h>

#include "helper_exc.h"

namespace vfd_euresys
{
	//////////////////
	// global const //
	//////////////////

	// max path lenthg used to keep vfs path
#define MAX_PATH 256
// max amount of caption/mask per video source
#define MAX_CAP  2
#define MAX_MASK 4
#define MAX_AUDIO_BUFFER 16384
	/////////////////
	// global enum //
	/////////////////


	// type of compressed frame
	enum FRAMETYPE {SPS,PPS,IDR, PS};
	// available color formats
	enum COLORFORMAT{RGB15, RGB16, RGB24, RGB32, YUV422, Y8, YUV420PL, YUV422PL} ;
	// available audio formats
	enum AUDIOFORMAT{G711A, G711U, PCM16, NOTDEF};
	////////////
	// struct //
	////////////

	// structure holding compressed data
	struct ENCODED_DATA
	{
		int index;
		char stamp[32];
		char file[MAX_PATH];
		int size;
		bool presence;
		FRAMETYPE type;
		char * data;
		bool reserved;

	};
	struct AUDIO_DATA
	{
		int index;
		char stamp[32];
		char file[MAX_PATH];
		int size;
		char * data;
		bool reserved;
	};

	struct FORMATTED_DATA
	{
		int index;
		char stamp[32];
		char file[MAX_PATH];
		bool presence;
		char * data;
		int fd;
	};



	//////////////////////
	// global functions //
	//////////////////////

	// function to browse to the next xml node
	xmlNodePtr GetNextNode (xmlNodePtr node)
	{
		xmlNodePtr cur;
		cur = node->next;
		if (cur->type == XML_TEXT_NODE)
			cur = cur->next;

		return cur;
	};

	// function to browse to the next child node
	xmlNodePtr GetNextChild (xmlNodePtr node)
	{
		xmlNodePtr cur;
		cur = node->xmlChildrenNode;;
		if (cur->type == XML_TEXT_NODE)
			cur = cur->next;

		return cur;
	};

	// create an xml file with a given header element title
	xmlTextWriterPtr CreateXMLFile(const char * filename, const char * title)
	{
		int ret;
		xmlTextWriterPtr writer;

		writer = xmlNewTextWriterFilename(filename, 0);
		if (writer == NULL)
			throw vfd_Exception(title , 105);// Can not open file for write operation

		// start the document with the xml default
		ret = xmlTextWriterStartDocument(writer, NULL, "UTF-8", NULL);
		if (ret < 0)
			throw vfd_Exception(title, 106);// write header

		// start Visual source element
		ret = xmlTextWriterStartElement(writer, BAD_CAST title);
		if (ret < 0)
			throw vfd_Exception(title, 107);// write element

		return writer;
	}


	///////////
	// class //
	///////////

	// clase managing a given source and source features
	class VisualSource
	{
	private:
		// path the the source folder
		char _SourcePath[MAX_PATH];
	protected:
		bool _configured;
		void CheckNode(xmlNodePtr node,xmlDocPtr doc)
        {
        	if (node == NULL)
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("VisualSource", 110);// no node
			}

        }


	public:
		// Path is the mount point of the board to be used
		// connector is the source connector
		VisualSource(const char * Path, int connector)
		{
			int i;

			// save path to the mounting location
			if (MAX_PATH < strlen(Path))
				throw vfd_Exception("VisualSource", 100);

			sprintf(_SourcePath,"%s/visual/source%d",Path, connector);

			// read elements value of config.xml
			ReadElements();

			// set default values for caption/mask
			for(i=0; i< MAX_CAP; i++)
			{
				m_Caption[i].text[0]= 0x00;
				m_Caption[i].pos = POSNOTDEF;
				m_Caption[i].X=0;
				m_Caption[i].Y=0;
			}

			for (i=0; i< MAX_MASK; i++)
			{
				m_Mask[i].X=0;
				m_Mask[i].Y=0;
				m_Mask[i].W=0;
				m_Mask[i].H=0;

			}
			// system is not ready to be used yet
			_configured = false;
		};

		~VisualSource(){ };

		// element defined as in the config.xml file
		bool m_VideoPresent;
		enum VIDEOSTANDARD {PAL, NTSC, CCIR, EIA, VSNOTDEF } m_VideoStandard;
		int m_Contrast;
		int m_Brightness;
		int m_Saturation;
		enum POSITION { TOPLEFT, BOTTOMLEFT, TOPRIGHT, BOTTOMRIGHT, POSNOTDEF};
		struct CAPTION
		{
			char text[47] ;
			POSITION pos;
			uint X;
			uint Y;
		} m_Caption[2];

		struct MASK
		{
			uint X;
			uint Y;
			uint W;
			uint H;
		} m_Mask[4];


		// read in the visual source filter configuration
		void ReadElements()
		{
			xmlChar * ctmp;

			//build config.xml full path
			char filename[MAX_PATH+12];
			sprintf(filename,"%s/config.xml", _SourcePath);

			// xml document & node pointer
			xmlDocPtr doc;
			xmlNodePtr cur;

			// read the file
			doc = xmlParseFile(filename);
			if (doc == NULL)
				throw vfd_Exception("VisualSource", 101); // no file

			// get file type
			cur = xmlDocGetRootElement(doc);
			if (cur == NULL)
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("VisualSource", 102);// no node
			}
			if (xmlStrcmp(cur->name, BAD_CAST "VisualSource"))
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("VisualSource", 103);// wrong type
			}

			// parse elements
			// verbose/concise option
			cur = GetNextChild(cur);
            CheckNode(cur,doc);

			if ( xmlStrcmp(cur->name, BAD_CAST "Verbose"))
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("VisualSource", 104);// Not verbose mode
			}

			// video presence
			cur = GetNextNode(cur);
            CheckNode(cur,doc);

			if ( ! xmlStrcmp(cur->name, BAD_CAST "VideoPresent"))
			{
				m_VideoPresent = true;
			}
			else
			{
				m_VideoPresent = false;
				// just stop here program will not work without video
				xmlFreeDoc(doc);
				throw vfd_Exception("VisualSource", 109);// No video detected

			}

			// detected video
			cur = GetNextChild(cur);
            CheckNode(cur,doc);

			m_VideoStandard = VSNOTDEF;
			if (!xmlStrcmp(cur->name, BAD_CAST "DetectedVideoStandard_PAL")) m_VideoStandard = PAL;
			if (!xmlStrcmp(cur->name, BAD_CAST "DetectedVideoStandard_NTSC")) m_VideoStandard = NTSC;
			if (!xmlStrcmp(cur->name, BAD_CAST "DetectedVideoStandard_CCIR")) m_VideoStandard = CCIR;
			if (!xmlStrcmp(cur->name, BAD_CAST "DetectedVideoStandard_EIA")) m_VideoStandard = EIA;
			// skip next element (video standard)
			//go one up
			cur = cur->parent;
			cur = GetNextNode(cur);
            CheckNode(cur,doc);
			//deinterlacingFilter
			cur = GetNextNode(cur);
            CheckNode(cur,doc);
			//contrast
			cur = GetNextNode(cur);
            CheckNode(cur,doc);

			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sscanf((const char*)ctmp,"%d",&m_Contrast);
			xmlFree(ctmp);

			// brightness
			cur = GetNextNode(cur);
            CheckNode(cur,doc);

			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sscanf((const char*)ctmp,"%d",&m_Brightness);
			xmlFree(ctmp);

			// saturation
			cur = GetNextNode(cur);
            CheckNode(cur,doc);

			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sscanf((const char*)ctmp,"%d",&m_Saturation);
			xmlFree(ctmp);

			xmlFreeDoc(doc);
		}



		// write the visual source filter configuration
		void WriteElements()
		{
			int ret;
			xmlTextWriterPtr writer;
			char tmp[64];

			// build config.xml full path
			char filename[MAX_PATH+12];
			sprintf(filename,"%s/config.xml", _SourcePath);

			// create a new XmlWriter for config.xml, with no compression.
			writer = CreateXMLFile(filename, "VisualSource");

			// start Video Standard element
			_configured = true;
			switch ( m_VideoStandard)
			{
			case PAL: ret = xmlTextWriterStartElement(writer, BAD_CAST "VideoStandard_PAL");
				break;
			case NTSC : ret = xmlTextWriterStartElement(writer, BAD_CAST "VideoStandard_NTSC");
				break;
			case CCIR : ret = xmlTextWriterStartElement(writer, BAD_CAST "VideoStandard_CCIR");
				break;
			case EIA : ret = xmlTextWriterStartElement(writer, BAD_CAST "VideoStandard_EIA");
				break;
			default : ret =1; _configured = false; break;

			}
			// write video standard ( no board is ready to be used
			if (ret == 0) xmlTextWriterEndElement(writer);
			if (ret < 0)
				throw vfd_Exception("VisualSource",107);// write element

			//write contrast, brightness, saturation
			ret = xmlTextWriterWriteFormatElement(writer, BAD_CAST "Contrast", "%d", m_Contrast);
			if (ret < 0)
				throw vfd_Exception("VisualSource", 107);// write element

			ret = xmlTextWriterWriteFormatElement(writer, BAD_CAST "Brightness", "%d", m_Brightness);
			if (ret < 0)
				throw vfd_Exception("VisualSource", 107);// write element

			ret = xmlTextWriterWriteFormatElement(writer, BAD_CAST "Saturation", "%d", m_Saturation);
			if (ret < 0)
				throw vfd_Exception("VisualSource", 107);// write element

			int i;
			// write all caption /mask data if available
			for (i=0; i< MAX_CAP; i++)
			{
				if (m_Caption[i].text)
				{
					sprintf(tmp,"Caption%d",i);
					// start caption description
					ret = xmlTextWriterStartElement(writer, BAD_CAST tmp);
					if (ret < 0)
						throw vfd_Exception("VisualSource", 107);// write element

					// caption Data
					ret = xmlTextWriterWriteFormatElement(writer, BAD_CAST "String", "%s", m_Caption[i].text);
					if (ret < 0)
						throw vfd_Exception("VisualSource", 107);// write element

					switch ( m_Caption[i].pos)
					{
					case TOPLEFT: ret = xmlTextWriterStartElement(writer, BAD_CAST "TopLeft");
						break;
					case BOTTOMLEFT : ret = xmlTextWriterStartElement(writer, BAD_CAST "BottomLeft");
						break;
					case TOPRIGHT : ret = xmlTextWriterStartElement(writer, BAD_CAST "TopRight");
						break;
					case BOTTOMRIGHT : ret = xmlTextWriterStartElement(writer, BAD_CAST "BottomRight");
						break;
					default :
						ret = xmlTextWriterWriteFormatElement(writer, BAD_CAST "X", "%d", m_Caption[i].X);
						if (ret < 0)
							throw vfd_Exception("VisualSource", 107);// write element

						ret = xmlTextWriterWriteFormatElement(writer, BAD_CAST "Y", "%d", m_Caption[i].Y);
						if (ret < 0)
							throw vfd_Exception("VisualSource", 107);// write element
						ret = 1;
						break;

					}
					if (ret == 0) xmlTextWriterEndElement(writer);

					if (ret < 0)
						throw vfd_Exception("VisualSource", 107);// write element


					// End Caption
					ret = xmlTextWriterEndElement(writer);
					if (ret < 0)
						throw vfd_Exception("VisualSource", 107);// write element

				}
			}
			for (i=0; i< MAX_MASK; i++)
			{
				if (m_Mask[i].W != 0 || m_Mask[i].H !=0)
				{
					sprintf(tmp,"Mask%d",i);
					// start mask description
					ret = xmlTextWriterStartElement(writer, BAD_CAST tmp);
					if (ret < 0)
						throw vfd_Exception("VisualSource", 107);// write element

					// mask data
					ret = xmlTextWriterWriteFormatElement(writer, BAD_CAST "X", "%d", m_Mask[i].X);
					if (ret < 0)
						throw vfd_Exception("VisualSource", 107);// write element

					ret = xmlTextWriterWriteFormatElement(writer, BAD_CAST "Y", "%d", m_Mask[i].Y);
					if (ret < 0)
						throw vfd_Exception("VisualSource", 107);// write element

					ret = xmlTextWriterWriteFormatElement(writer, BAD_CAST "W", "%d", m_Mask[i].W);
					if (ret < 0)
						throw vfd_Exception("VisualSource", 107);// write element

					ret = xmlTextWriterWriteFormatElement(writer, BAD_CAST "H", "%d", m_Mask[i].H);
					if (ret < 0)
						throw vfd_Exception("VisualSource", 107);// write element


					// end mask
					ret = xmlTextWriterEndElement(writer);
					if (ret < 0)
						throw vfd_Exception("VisualSource", 107);// write element

				}
			}

			// close Visual Source
			ret = xmlTextWriterEndElement(writer);
			if (ret < 0)
				throw vfd_Exception("VisualSource", 107);// write element

			// close document
			ret = xmlTextWriterEndDocument(writer);
			if (ret < 0)
				throw vfd_Exception("VisualSource", 108);// close file

			xmlFreeTextWriter(writer);
		}

	};

	// class managing compress data stream
	class VisualEncoder: public VisualSource
	{
    protected:
		void CheckNode(xmlNodePtr node,xmlDocPtr doc)
        {
        	if (node == NULL)
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("VisualEncoder", 110);// no node
			}

        }


	private:
		char _EncoderPath[MAX_PATH];
		char _EnConfigPath[MAX_PATH];
		char _MetaPath[MAX_PATH];
		ENCODED_DATA _Data;
		int _SizeX;
		int _SizeY;
		int _BufferSize;

		// write a node and its value in the config file
		void WriteData(const char * Name, int Value)
		{
			int ret;
			xmlTextWriterPtr writer;

			// create a new XmlWriter for config.xml, with no compression.
			writer = CreateXMLFile(_EnConfigPath, "VisualEncoder");

			// write node and value
			ret = xmlTextWriterWriteFormatElement(writer, BAD_CAST Name, "%d", Value);
			if (ret < 0)
				throw vfd_Exception("VisualEncoder",107);// write element

			// close Visual Encoder
			ret = xmlTextWriterEndElement(writer);
			if (ret < 0)
				throw vfd_Exception("VisualEncoder",107);// write element

			// close document
			ret = xmlTextWriterEndDocument(writer);
			if (ret < 0)
				throw vfd_Exception("VisualEncoder",108);// cose file

			xmlFreeTextWriter(writer);

		}
		// write an empty node in the config file
		void WriteData(const char * Name)
		{
			int ret;
			xmlTextWriterPtr writer;

			// create a new XmlWriter for config.xml, with no compression.
			writer = CreateXMLFile(_EnConfigPath, "VisualEncoder");

			// write node
			ret = xmlTextWriterStartElement(writer, BAD_CAST Name);
			if (ret < 0)
				throw vfd_Exception("VisualEncoder", 107);// write element

			// close element
			ret = xmlTextWriterEndElement(writer);
			if (ret < 0)
				throw vfd_Exception("VisualEncoder", 107);// write element

			// close Visual Encoder
			ret = xmlTextWriterEndElement(writer);
			if (ret < 0)
				throw vfd_Exception("VisualEncoder", 107);// write element

			// close document
			ret = xmlTextWriterEndDocument(writer);
			if (ret < 0)
				throw vfd_Exception("VisualEncoder", 108);// close file

			xmlFreeTextWriter(writer);

		};
		// read data from the metadata.xml file return the amount of available data chunks
		int ReadMetaData()
		{
			xmlChar * ctmp;
			int count;
			// xml document & node pointer
			xmlDocPtr doc;
			xmlNodePtr cur;

			// read the file
			doc = xmlParseFile(_MetaPath);

			if (doc == NULL)
				throw vfd_Exception("VisualEncoder", 202); // no file

			// get file type
			cur = xmlDocGetRootElement(doc);
			if (cur == NULL)
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("VisualEncoder", 203);// no node
			}

			if (xmlStrcmp(cur->name, BAD_CAST "VisualEncoderMetaDataFile"))
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("VisualEncoder", 204);// wrong type
			}

			// parse elements
			// verbose/concise option
			cur = GetNextChild(cur);
			CheckNode(cur,doc);
			if ( xmlStrcmp(cur->name, BAD_CAST "Verbose"))
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("VisualEncoder", 104);// not verbose mode
			}

			// data Count
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sscanf((const char*)ctmp,"%d",&count);
			xmlFree(ctmp);

			// file is empty
			if (!count)
			{
				// finish with this metadata
				xmlFreeDoc(doc);
				return 0;
			}

			// file has some metadata, get them
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			cur = GetNextChild(cur);
			CheckNode(cur,doc);
			// index
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sscanf((const char*)ctmp,"%d",&_Data.index);
			xmlFree(ctmp);

			// timestamp
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			strncpy (_Data.stamp, (const char*) ctmp,  32);
			xmlFree(ctmp);

			// file name
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sprintf(_Data.file,"%sdata/%s",_EncoderPath, (const char*) ctmp);
			xmlFree(ctmp);

			// size
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sscanf((const char*)ctmp,"%d",&_Data.size);
			xmlFree(ctmp);

			// video presence
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			if ( ! xmlStrcmp(cur->name, BAD_CAST "VideoPresent"))
			{
				_Data.presence = true;
			}
			else
			{
				_Data.presence = false;
			}

			// slice type
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			if (!xmlStrcmp(cur->name, BAD_CAST "IDRSlice")) _Data.type = IDR;
			if (!xmlStrcmp(cur->name, BAD_CAST "PSlice")) _Data.type = PS;
			if (!xmlStrcmp(cur->name, BAD_CAST "SequenceParameterSet")) _Data.type = SPS;
			if (!xmlStrcmp(cur->name, BAD_CAST "PictureParameterSet")) _Data.type = PPS;

			// finish with this metadata
			xmlFreeDoc(doc);

			return count;

		}
		//read data from config.xml file
		int ReadData()
		{
			xmlChar * ctmp;
			int count=0;
			// xml document & node pointer
			xmlDocPtr doc;
			xmlNodePtr cur;

			// read the file
			doc = xmlParseFile(_EnConfigPath);

			if (doc == NULL)
				throw vfd_Exception("VisualEncoder", 202); // no file

			// get file type
			cur = xmlDocGetRootElement(doc);
			if (cur == NULL)
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("VisualEncoder", 203);// no node
			}

			if (xmlStrcmp(cur->name, BAD_CAST "VisualEncoder"))
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("VisualEncoder", 204);// wrong type
			}

			// verbose/concise option
			cur = GetNextChild(cur);
			CheckNode(cur,doc);
			if ( xmlStrcmp(cur->name, BAD_CAST "Verbose"))
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("VisualEncoder", 104);// not verbose mode
			}

			// parse elements
			// active/idle
			cur = GetNextNode(cur);
            CheckNode(cur,doc);
			// decimation
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			// compression type
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			// gop
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			// scaling
			cur = GetNextNode(cur);
			CheckNode(cur,doc);

			// sizeX
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sscanf((const char*)ctmp,"%d",&_SizeX);
			xmlFree(ctmp);

			// sizeY
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sscanf((const char*)ctmp,"%d",&_SizeY);
			xmlFree(ctmp);

			// buffer Size
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sscanf((const char*)ctmp,"%d",&_BufferSize);
			xmlFree(ctmp);

			// finish with this metadata
			xmlFreeDoc(doc);

			return count;

		}

	public:
		// Path is the mount point of the board to be used
		// connector is the source connector
		VisualEncoder(const char * Path, int connector):VisualSource(Path, connector)
		{
			// build internally used folder path
			sprintf(_EncoderPath,"%s/visual/source%d/encoder/",Path, connector);
			sprintf(_EnConfigPath,"%s/visual/source%d/encoder/config.xml",Path, connector);
			sprintf(_MetaPath,"%s/visual/source%d/encoder/metadata.xml",Path, connector);
			_Data.reserved = false;
			_Data.data =NULL;
			ReadData();
			// pre allocate buffer
			_Data.data = (char*) malloc (_BufferSize);

			if (!_Data.data) throw vfd_Exception("VisualEncoder", 210);// out of memory
		}
		~VisualEncoder()
		{
			_EncoderPath[0]=0x00;
			_EnConfigPath[0]=0x00;
			_MetaPath[0]=0x00;
			if (_Data.data) free (_Data.data);
		}

		void StartAcquisition()
		{
			// if the system is ready , write Active in the config.xml
			if (_configured)
				WriteData("Active");
			else
				throw vfd_Exception("VisualEncoder", 201);// not configured
		}

		void StopAcquisition()
		{
			WriteData("Idle");
		}

		int GetSizeX()
		{
			return _SizeX;
		}

		int GetSizeY()
		{
			return _SizeY;
		}

		int GetBufferSize()
		{
			return _BufferSize;
		}


		void SetBufferCount(int Count)
		{
			// write data to config.xml file
			WriteData("BufferCount",Count);

		}

		// return H264 compress data ( blocking, using a I/O poll system)
		ENCODED_DATA GetData()
		{
			int datacount = 0;
			pollfd fdinfo;
			int fd;
			// check if live and buffer released
			if (_Data.reserved)
				return _Data;
			_Data.reserved = true;

			if (!_configured)
				throw vfd_Exception("VisualEncoder", 201);// Not configured

			// get meta data count
			while(!datacount)
			{
				datacount = ReadMetaData();
				if (! datacount)
				{
					// data not available => wait
					fd = open ( _MetaPath,O_RDONLY);
					fdinfo.fd = fd;
					fdinfo.events = POLLIN;
					// register a poll on the file to get wackup when data are available
					// this call is blocking
					poll(&fdinfo,1,1000);
					close(fd);
				}
			}
			// data available => read data from file
			fd = open ( _Data.file,O_RDONLY);
			if (fd <0)
				throw vfd_Exception("VisualEncoder", 209);// no access to data
			read(fd, _Data.data, _Data.size);
			close (fd);

			return _Data;
		};

		// let the driver know that the data a ready to be re-used
		void ReleaseData()
		{
			// check if buffer is to be released
			if (!_Data.reserved)
				return; // nothing to release
			_Data.reserved = false;

			// unlink the file and the driver will know that the file can be re-used
			unlink(_Data.file);
		};

	};
	// classe managing raw data stre
	class VisualFormatter: public VisualSource
	{
    protected:
		void CheckNode(xmlNodePtr node,xmlDocPtr doc)
        {
        	if (node == NULL)
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("VisualFormatter", 110);// no node
			}

        }

	private:
		char _FormatterPath[MAX_PATH];
		char _FrConfigPath[MAX_PATH];
		char _MetaPath[MAX_PATH];
		FORMATTED_DATA * _pData;
		int _SizeX;
		int _SizeY;
		int _BufferSize;
		int _BufferCount;
		// write a node and its value in the config file
		void WriteData(const char * Name, int Value)
		{
			int ret;
			xmlTextWriterPtr writer;

			// create a new XmlWriter for config.xml, with no compression.
			writer = CreateXMLFile(_FrConfigPath, "VisualFormatter");

			// write node and value
			ret = xmlTextWriterWriteFormatElement(writer, BAD_CAST Name, "%d", Value);
			if (ret < 0)
				throw vfd_Exception("VisualFormatter", 107);// write element

			// close Visual Formatter
			ret = xmlTextWriterEndElement(writer);
			if (ret < 0)
				throw vfd_Exception("VisualFormatter", 107);// write element

			// close document
			ret = xmlTextWriterEndDocument(writer);
			if (ret < 0)
				throw vfd_Exception("VisualFormatter", 108);// Close file

			xmlFreeTextWriter(writer);

		}
		// write an empty node in the config file
		void WriteData(const char * Name)
		{
			int ret;
			xmlTextWriterPtr writer;

			// create a new XmlWriter for config.xml, with no compression.
			writer = CreateXMLFile(_FrConfigPath, "VisualFormatter");

			// write node
			ret = xmlTextWriterStartElement(writer, BAD_CAST Name);
			if (ret < 0)
				throw vfd_Exception("VisualFormatter", 107);// write element

			// close element
			ret = xmlTextWriterEndElement(writer);
			if (ret < 0)
				throw vfd_Exception("VisualFormatter", 107);// write element

			// close Visual Formatter
			ret = xmlTextWriterEndElement(writer);
			if (ret < 0)
				throw vfd_Exception("VisualFormatter", 107);// write element

			// close document
			ret = xmlTextWriterEndDocument(writer);
			if (ret < 0)
				throw vfd_Exception("VisualFormatter", 108);// close file

			xmlFreeTextWriter(writer);

		};


		// read data from the metadata.xml file return a data chunk / NULL is nothing is available
		FORMATTED_DATA * ReadMetaData()
		{
			xmlChar * ctmp;
			int count;
			// xml document & node pointer
			xmlDocPtr doc;
			xmlNodePtr cur;

			// read the file
			doc = xmlParseFile(_MetaPath);

			if (doc == NULL)
				throw vfd_Exception("VisualFormatter", 202); // no file

			// get file type
			cur = xmlDocGetRootElement(doc);
			if (cur == NULL)
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("VisualFormatter", 203);// no node
			}

			if (xmlStrcmp(cur->name, BAD_CAST "VisualFormatterMetaDataFile"))
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("VisualFormatter", 204);// wrong type
			}

			// parse elements
			// verbose/concise option
			cur = GetNextChild(cur);
			CheckNode(cur,doc);
			if ( xmlStrcmp(cur->name, BAD_CAST "Verbose"))
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("VisualFormatter", 104);// Not verbose mode
			}

			// data count
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sscanf((const char*)ctmp,"%d",&count);
			xmlFree(ctmp);
			// file is empty
			if (!count)
			{
				// finish with this metadata
				xmlFreeDoc(doc);
				return NULL;
			}

			// file has some metadata, get them
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			cur = GetNextChild(cur);
			CheckNode(cur,doc);
			// index
			int index;
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sscanf((const char*)ctmp,"%d",&index);
			xmlFree(ctmp);

			// time stamp
			char stamp[32];
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			strncpy (stamp, (const char*) ctmp,  32);
			xmlFree(ctmp);

			// file
			int fn;
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sscanf((const char*)ctmp,"%d",&fn);

			// mmap file if this is not done yet
			if (!_pData[fn].fd)
			{
				// build path to file
				sprintf(_pData[fn].file,"%sdata/%s",_FormatterPath, (const char*) ctmp);
				// open the file
				_pData[fn].fd = open(_pData[fn].file, O_RDONLY);
				if (_pData[fn].fd <0)
					throw vfd_Exception("VisualFormatter", 309);// no access to data
				// map the file in virtual memory space
				_pData[fn].data = (char*) mmap((caddr_t) 0 , _BufferSize, PROT_READ, MAP_SHARED, _pData[fn].fd, 0);
				if(_pData[fn].data == MAP_FAILED)
				{
					close (_pData[fn].fd);
					throw vfd_Exception("VisualFormatter", 310);// mapping failed
				}
			}
			// store pending data
			_pData[fn].index = index;
			strncpy(_pData[fn].stamp, stamp, 32);
			xmlFree(ctmp);

			// size
			cur = GetNextNode(cur);
            CheckNode(cur,doc);
			// video presence
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			if ( ! xmlStrcmp(cur->name, BAD_CAST"VideoPresent"))
			{
				_pData[fn].presence = true;
			}
			else
			{
				_pData[fn].presence = false;
			}

			// finish with this metadata
			xmlFreeDoc(doc);

			return & _pData[fn];

		}

		//read data from config.xml file
		int ReadData()
		{
			xmlChar * ctmp;
			int count;
			// xml document & node pointer
			xmlDocPtr doc;
			xmlNodePtr cur;

			// read the file
			doc = xmlParseFile(_FrConfigPath);

			if (doc == NULL)
				throw vfd_Exception("VisualFormatter", 202); // no file

			// get file type
			cur = xmlDocGetRootElement(doc);
			if (cur == NULL)
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("VisualFormatter", 203);// no node
			}

			if (xmlStrcmp(cur->name, BAD_CAST "VisualFormatter"))
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("VisualFormatter", 204);// wrong type
			}


			// verbose/concise option
			cur = GetNextChild(cur);
			CheckNode(cur,doc);
			if ( xmlStrcmp(cur->name, BAD_CAST "Verbose"))
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("VisualFormatter", 104);// Not verbose mode
			}
			// parse elements
			// active/idle
			cur = GetNextNode(cur);
            CheckNode(cur,doc);
			// decimation
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			// format type
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			// scaling
			cur = GetNextNode(cur);
            CheckNode(cur,doc);
			// sizeX
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sscanf((const char*)ctmp,"%d",&_SizeX);
			xmlFree(ctmp);

			// sizeY
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sscanf((const char*)ctmp,"%d",&_SizeY);
			xmlFree(ctmp);

			// buffer size
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sscanf((const char*)ctmp,"%d",&_BufferSize);
			xmlFree(ctmp);

			// finish with this metadata
			xmlFreeDoc(doc);

			return count;

		}

	public:
		// Path is the mount point of the board to be used
		// connector is the source connector
		VisualFormatter(const char * Path, int connector):VisualSource(Path, connector)
		{
			sprintf(_FormatterPath,"%s/visual/source%d/formatter/",Path, connector);
			sprintf(_FrConfigPath,"%s/visual/source%d/formatter/config.xml",Path, connector);
			sprintf(_MetaPath,"%s/visual/source%d/formatter/metadata.xml",Path, connector);
			ReadData();
			_BufferCount = 3; // default value
			_pData = (FORMATTED_DATA*) malloc (sizeof(FORMATTED_DATA)*_BufferCount);
			int i;
			for (i=0; i<_BufferCount; i++)
			{
				_pData[i].data = NULL;
				_pData[i].fd = 0;
			}
		}
		~VisualFormatter()
		{
			_FormatterPath[0]=0x00;
			_FrConfigPath[0]=0x00;
			_MetaPath[0]=0x00;
			int i;
			// un map close data file if needed
			for (i=0; i<_BufferCount; i++)
			{
				if (_pData[i].data) munmap(_pData[i].data, _BufferSize);
				if (_pData[i].fd) close (_pData[i].fd);
			}
			free (_pData);
		}


		void StartAcquisition()
		{
			if (_configured)
				WriteData("Active");
			else
				throw vfd_Exception("VisualFormatter", 201);// Not configured
		}

		void StopAcquisition()
		{
			WriteData("Idle");
			// un-map all links and close files
			int i;
			for (i=0; i<_BufferCount; i++)
			{
				if (_pData[i].data) munmap(_pData[i].data, _BufferSize);
				if (_pData[i].fd)
				{
					close (_pData[i].fd);
					_pData[i].fd = 0;
				}
			}

		}

		int GetSizeX()
		{
			return _SizeX;
		}

		int GetSizeY()
		{
			return _SizeY;
		}

		int GetBufferSize()
		{
			return _BufferSize;
		}

		// set the amount of buffer default is 3 (optionnal settings)
		void SetBufferCount(int Count)
		{
			// write data to config.xml file
			WriteData("BufferCount",Count);

			// unmap all links / close files
			int i;
			for (i=0; i<_BufferCount; i++)
			{
				if (_pData[i].data) munmap(_pData[i].data, _BufferSize);
				if (_pData[i].fd) close (_pData[i].fd);
			}
			free (_pData);

			// reallocate
			_BufferCount = Count; // default value
			_pData = (FORMATTED_DATA*) malloc (sizeof(FORMATTED_DATA)*_BufferCount);
			for (i=0; i<_BufferCount; i++)
			{
				_pData[i].data = NULL;
				_pData[i].fd = 0;
			}
		}
		// set Color format
		void SetColorFormat ( COLORFORMAT format)
		{
			char color[16];
			switch (format)
			{
			case RGB15: strcpy (color, "FormatRGB15");
				break;

			case RGB16: strcpy (color, "FormatRGB16");
				break;

			case RGB24: strcpy (color, "FormatRGB24");
				break;

			case RGB32: strcpy (color, "FormatRGB32");
				break;

			case YUV422: strcpy (color, "FormatYUV422");
				break;

			case Y8: strcpy (color, "FormatY8");
				break;

			case YUV420PL: strcpy (color, "FormatYUV420PL");
				break;

			case YUV422PL: strcpy (color, "FormatYUV422PL");
				break;
			default :
				throw vfd_Exception("VisualFormatter", 311);// unknown color format
			}

			WriteData(color);
			ReadData();

		}
		// return raw picture ( blocking, using a I/O poll system)

		FORMATTED_DATA GetData()
		{
			FORMATTED_DATA * pdata = NULL;
			pollfd fdinfo;
			int fd;

			if (!_configured)
				throw vfd_Exception("VisualFormatter", 201);// Not configured

			// get meta data count
			while(!pdata)
			{
				pdata = ReadMetaData();
				if (! pdata)
				{
					// data not available => wait
					fd = open ( _MetaPath,O_RDONLY);
					fdinfo.fd = fd;
					fdinfo.events = POLLIN;
					// register a poll on the file to get wackup when data are available
					poll(&fdinfo,1,1000);
					close(fd);
				}
			}
			return pdata[0];
		};

		// let the driver know that the data a ready to be re-used
		void ReleaseData(FORMATTED_DATA data)
		{
			unlink(data.file);
		}
	};

	// clase managing a given audio source and its features
	class AudioSource
	{
    protected:
		void CheckNode(xmlNodePtr node,xmlDocPtr doc)
        {
        	if (node == NULL)
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("AudioSource", 110);// no node
			}

        }

	private:
		// path the the source folder
		char _SourcePath[MAX_PATH];
		char _DataPath[MAX_PATH];
		char _ConfigPath[MAX_PATH];
		char _MetaPath[MAX_PATH];

		// element defined as in the config.xml file
		int m_Volume;
		AUDIOFORMAT m_Format;
		int m_SamplingRate;
		int m_BufferCount;
		AUDIO_DATA _Data;
		// read data from the metadata.xml file return the amount of available data chunks
		int ReadMetaData()
		{
			xmlChar * ctmp;
			int count;
			// xml document & node pointer
			xmlDocPtr doc;
			xmlNodePtr cur;

			// read the file
			doc = xmlParseFile(_MetaPath);

			if (doc == NULL)
				throw vfd_Exception("AudioSource", 202); // no file

			// get file type
			cur = xmlDocGetRootElement(doc);
			if (cur == NULL)
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("AudioSource", 203);// no node
			}

			if (xmlStrcmp(cur->name, BAD_CAST "AudioEncoderMetaDataFile"))
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("AudioSource", 204);// wrong type
			}

			// parse elements
			// verbose/concise option
			cur = GetNextChild(cur);
			CheckNode(cur,doc);
			if ( xmlStrcmp(cur->name, BAD_CAST "Verbose"))
			{
				xmlFreeDoc(doc);
				throw vfd_Exception("AudioSource", 104);// not verbose mode
			}

			// data Count
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sscanf((const char*)ctmp,"%d",&count);
			xmlFree(ctmp);

			// file is empty
			if (!count)
			{
				// finish with this metadata
				xmlFreeDoc(doc);
				return 0;
			}

			// file has some metadata, get them
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			cur = GetNextChild(cur);
			CheckNode(cur,doc);
			// index
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sscanf((const char*)ctmp,"%d",&_Data.index);
			xmlFree(ctmp);

			// timestamp
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			strncpy (_Data.stamp, (const char*) ctmp,  32);
			xmlFree(ctmp);

			// file name
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sprintf(_Data.file,"%s/%s",_DataPath, (const char*) ctmp);
			xmlFree(ctmp);

			// size
			cur = GetNextNode(cur);
			CheckNode(cur,doc);
			ctmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			sscanf((const char*)ctmp,"%d",&_Data.size);
			xmlFree(ctmp);

			// finish with this metadata
			xmlFreeDoc(doc);

			return count;

		}

		// write the visual source filter configuration
		int WriteElements()
		{
			int ret;
			//create a  xml file for writting
			xmlTextWriterPtr writer = CreateXMLFile(_ConfigPath, "AudioEncoder");

			// write Volume
			ret = xmlTextWriterWriteFormatElement(writer, BAD_CAST "Volume", "%d", m_Volume);
			if (ret < 0)
				throw vfd_Exception("AudioSource", 107);// write element

						// start Audio Standard element
			if ( m_SamplingRate != 0 && m_Format != NOTDEF )
			{
				switch ( m_Format )
				{
					case G711A : ret = xmlTextWriterStartElement ( writer, BAD_CAST "FormatG711ALAW" );
						break;

					case G711U : ret = xmlTextWriterStartElement ( writer, BAD_CAST "FormatG711ULAW" );
						break;

					case PCM16 : ret = xmlTextWriterStartElement ( writer, BAD_CAST "FormatLPCM16" );
						switch ( m_SamplingRate )
						{
							case 8000 : xmlTextWriterStartElement ( writer, BAD_CAST "SamplingRate8KHz" );
								break;
							case 16000 : xmlTextWriterStartElement ( writer, BAD_CAST "SamplingRate16KHz" );
								break;
							case 22050 : xmlTextWriterStartElement ( writer, BAD_CAST "SamplingRate22.05KHz" );
								break;
							case 44100 : xmlTextWriterStartElement ( writer, BAD_CAST "SamplingRate44.1KHz" );
								break;
							case 48000 : xmlTextWriterStartElement ( writer, BAD_CAST "SamplingRate48KHz" );
								break;
							default : xmlTextWriterStartElement ( writer, BAD_CAST "SamplingRate8KHz" );
								break;
						}
						xmlTextWriterEndElement ( writer );
						break;
				}
				// end format element
				if ( ret == 0 ) xmlTextWriterEndElement ( writer );
				if ( ret < 0 )
					throw vfd_Exception ( "AudioSource",107 );// write element
			}
			//write Buffer count
			ret = xmlTextWriterWriteFormatElement(writer, BAD_CAST "BufferCount", "%d", m_BufferCount);
			if (ret < 0)
				throw vfd_Exception("AudioSource", 107);// write element


			// close Audio Source
			ret = xmlTextWriterEndElement(writer);
			if (ret < 0)
				throw vfd_Exception("AudioSource", 107);// write element

			// close document
			ret = xmlTextWriterEndDocument(writer);
			if (ret < 0)
				throw vfd_Exception("AudioSource", 108);// close file

			xmlFreeTextWriter(writer);
		}


		// write a node and its value in the config file
		void WriteData(const char * Name, int Value)
		{
			int ret;
			xmlTextWriterPtr writer;

			// create a new XmlWriter for config.xml, with no compression.
			writer = CreateXMLFile(_ConfigPath, "AudioEncoder");

			// write node and value
			ret = xmlTextWriterWriteFormatElement(writer, BAD_CAST Name, "%d", Value);
			if (ret < 0)
				throw vfd_Exception("AudioSource",107);// write element

			// close Visual Encoder
			ret = xmlTextWriterEndElement(writer);
			if (ret < 0)
				throw vfd_Exception("AudioSource",107);// write element

			// close document
			ret = xmlTextWriterEndDocument(writer);
			if (ret < 0)
				throw vfd_Exception("AudioSource",108);// cose file

			xmlFreeTextWriter(writer);

		}

		// write an empty node in the config file
		void WriteData(const char * Name)
		{
			int ret;
			xmlTextWriterPtr writer;

			// create a new XmlWriter for config.xml, with no compression.
			writer = CreateXMLFile(_ConfigPath, "AudioEncoder");

			// write node
			ret = xmlTextWriterStartElement(writer, BAD_CAST Name);
			if (ret < 0)
				throw vfd_Exception("AudioSource", 107);// write element

			// close element
			ret = xmlTextWriterEndElement(writer);
			if (ret < 0)
				throw vfd_Exception("AudioSource", 107);// write element

			// close Visual Encoder
			ret = xmlTextWriterEndElement(writer);
			if (ret < 0)
				throw vfd_Exception("AudioSource", 107);// write element

			// close document
			ret = xmlTextWriterEndDocument(writer);
			if (ret < 0)
				throw vfd_Exception("AudioSource", 108);// close file

			xmlFreeTextWriter(writer);

		};

	public:
		// Path is the mount point of the board to be used
		// connector is the source connector
		// format is the audio format
		// sampling is the sampling rate ( if format is LPCM16)
		AudioSource(const char * Path, int connector, AUDIOFORMAT format = NOTDEF, int sampling = 0)
		{
			// save path to the mounting location
			if (MAX_PATH < strlen(Path))
				throw vfd_Exception("AudioSource",100);
			sprintf(_SourcePath,"%s/audio/source%d",Path, connector);
			sprintf(_DataPath,"%s/data",_SourcePath);
			sprintf(_ConfigPath,"%s/config.xml",_SourcePath);
			sprintf(_MetaPath,"%s/metadata.xml",_SourcePath);

			m_Format = format;
			m_SamplingRate = sampling;
			m_BufferCount = 10;
			m_Volume = 100;
			// write data to config.xml
			WriteElements();
			_Data.reserved = false;
			// pre allocate buffer
			_Data.data = (char*) malloc (MAX_AUDIO_BUFFER);

			if (!_Data.data) throw vfd_Exception("AudioSource", 210);// out of memory

		};

		~AudioSource()
		{
			if (_Data.data) free (_Data.data);
		}
		void StartAcquisition()
		{
			WriteData("Active");
		}

		void StopAcquisition()
		{
			WriteData("Idle");
		}

		void SetVolume(int Volume)
		{
			// write data to config.xml file
			WriteData("Volume",Volume);

		}

		// return audio data ( blocking, using a I/O poll system)
		AUDIO_DATA GetData()
		{
			int datacount = 0;
			pollfd fdinfo;
			int fd;
			// check if live and buffer released
			if (_Data.reserved)
				return _Data;
			_Data.reserved = true;

			// get meta data count
			while(!datacount)
			{
				datacount = ReadMetaData();
				if (! datacount)
				{
					// data not available => wait
					fd = open ( _MetaPath,O_RDONLY);
					fdinfo.fd = fd;
					fdinfo.events = POLLIN;
					// register a poll on the file to get wackup when data are available
					// this call is blocking
					poll(&fdinfo,1,1000);
					close(fd);
				}
			}
			// data available => read data from file
			fd = open ( _Data.file,O_RDONLY);
			if (fd <0)
				throw vfd_Exception("AudioSource", 209);// no access to data
			read(fd, _Data.data, _Data.size);
			close (fd);

			return _Data;
		};

		// let the driver know that the data a ready to be re-used
		void ReleaseData()
		{
			// check if buffer is to be released
			if (!_Data.reserved)
				return; // nothing to release
			_Data.reserved = false;

			// unlink the file and the driver will know that the file can be re-used
			unlink(_Data.file);
		};

	};

};

