/*
+-------------------------------- DISCLAIMER ---------------------------------+
|                                                                             |
| This application program is provided to you free of charge as an example.   |
| Despite the considerable efforts of Euresys personnel to create a usable    |
| example, you should not assume that this program is error-free or suitable  |
| for any purpose whatsoever.                                                 |
|                                                                             |
| EURESYS does not give any representation, warranty or undertaking that this |
| program is free of any defect or error or suitable for any purpose. EURESYS |
| shall not be liable, in contract, in torts or otherwise, for any damages,   |
| loss, costs, expenses or other claims for compensation, including those     |
| asserted by third parties, arising out of or in connection with the use of  |
| this program.                                                               |
|                                                                             |
+-----------------------------------------------------------------------------+
*/

// 2.0.0.15

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <iostream>
#include <stdlib.h>

#include "SDL.h"
#include "../Common/helper_vfs.h"
#include "../Common/helper_avc.h"

using namespace std;
using namespace vfd_euresys;
bool CheckForQuit()
{
	SDL_Event event;

	// poll for events, and handle <ESC> key
	while (SDL_PollEvent(&event)) 
	{
		switch (event.type) 
		{
		case SDL_KEYUP:
			// If escape is pressed, return (and thus, quit)
			if (event.key.keysym.sym == SDLK_ESCAPE)
				return true;
			break;
		case SDL_QUIT:
			return true;
		}
	}

	return false;
}

void CopyPictureToOverlay(AVFrame *picture, SDL_Overlay *overlay)
{
	uint8_t *overlayData;
	uint8_t *pictureData;

	SDL_LockYUVOverlay(overlay);
	for (int plane = 0; plane < overlay->planes; plane++) {
		overlayData = overlay->pixels[plane];
		pictureData = picture->data[plane];
		int lineSize = overlay->pitches[plane] < picture->linesize[plane] ? 
			overlay->pitches[plane] : picture->linesize[plane];
		for (int line = 0; line < overlay->h; line++) {
			memcpy(overlayData, pictureData, lineSize);
			overlayData += overlay->pitches[plane];
			pictureData += picture->linesize[plane];
		}
	}
	SDL_UnlockYUVOverlay(overlay);
}

int CreateSdlComponents(SDL_Overlay *&overlay, SDL_Rect &rect, 
	SDL_Surface *&screen, VisualEncoder * source)
{
	// Create a width * height window with 32 bits per pixels.
	screen = SDL_SetVideoMode(source->GetSizeX(),source->GetSizeY(),
		32, SDL_SWSURFACE | SDL_DOUBLEBUF);
	if (screen == NULL) {
		std::cerr << "Unable to set video mode: " << SDL_GetError() << std::endl;
		return -1;
	}
	overlay = SDL_CreateYUVOverlay(source->GetSizeX(), source->GetSizeY(), 
		SDL_IYUV_OVERLAY, screen);
	if (overlay == NULL) {
		std::cerr << "Unable to create overlay" << std::endl;
		return -1;
	}
	rect.x = 0;
	rect.y = 0;
	rect.w = source->GetSizeX();
	rect.h = source->GetSizeY();
	return 0;
}


void RenderH264Frames(VisualEncoder * source)
{
	SDL_Overlay *overlay;
	SDL_Rect rect;
	SDL_Surface *screen;
	AvcodecH264Decoder decoder;

	int status = CreateSdlComponents(overlay, rect, screen, source);
	if (status != 0) {
		return;
	}

	while (!CheckForQuit()) {
		ENCODED_DATA Data = source->GetData();
		AVFrame *picture = decoder.Decode(Data.data, Data.size);
		if (picture) {
			CopyPictureToOverlay(picture, overlay);
			int status = SDL_DisplayYUVOverlay(overlay, &rect);
			if (status != 0) {
				std::cerr << "Unable to display the overlay" << std::endl;
				SDL_FreeYUVOverlay(overlay);
				return;
			}
		}
		source->ReleaseData();
	}
	SDL_FreeYUVOverlay(overlay);
}

void RenderH264Video(VisualEncoder * source)
{
	source->StopAcquisition();
	source->StartAcquisition();
	RenderH264Frames(source);
	source->StopAcquisition();
}



int main(int argc, char *argv[])
{
	int status = 0;

	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		std::cerr << "Unable to init SDL: " << SDL_GetError() << std::endl;
		return -1;
	}

	try 
	{    
		VisualEncoder mysource("/mnt/u4_0", 0);
		mysource.SetBufferCount(10);
		sprintf(mysource.m_Caption[0].text, "vid %d",1);
		mysource.m_Caption[0].pos = VisualSource::BOTTOMLEFT;

		sprintf(mysource.m_Caption[1].text, "H.264 Video");
		mysource.m_Caption[1].X = 50;
		mysource.m_Caption[1].Y = 50;

		mysource.m_Mask[0].X = 50;
		mysource.m_Mask[0].Y = 85;
		mysource.m_Mask[0].W = 180;
		mysource.m_Mask[0].H = 10;
		mysource.WriteElements();

		RenderH264Video(&mysource);
	} 
	catch (std::exception &e)
	{
		std::cerr << "--- EXCEPTION --- " << std::endl;
		std::cerr << e.what() << std::endl;
		std::cerr << std::endl;
		status = -1;
	}
	catch (vfd_euresys::vfd_Exception &e)
	{
		std::cerr << "--- EXCEPTION in VFD --- " << std::endl;
		std::cerr << e.what() << std::endl;
		std::cerr << std::endl;
		status = -1;
	}
	SDL_Quit();

	return status;
};
